@@ -7,11 +7,11 @@
 ** This file is part of the tools applications of the Qt Toolkit.
 **
 ** $QT_BEGIN_LICENSE:LGPL$
-** No Commercial Usage
-** This file contains pre-release code and may not be distributed.
-** You may use this file in accordance with the terms and conditions
-** contained in the Technology Preview License Agreement accompanying
-** this package.
+** Commercial Usage
+** Licensees holding valid Qt Commercial licenses may use this file in
+** accordance with the Qt Commercial License Agreement provided with the
+** Software or, alternatively, in accordance with the terms contained in
+** a written agreement between you and Nokia.
 **
 ** GNU Lesser General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU Lesser
@@ -25,16 +25,16 @@
 ** rights.  These rights are described in the Nokia Qt LGPL Exception
 ** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
 **
+** GNU General Public License Usage
+** Alternatively, this file may be used under the terms of the GNU
+** General Public License version 3.0 as published by the Free Software
+** Foundation and appearing in the file LICENSE.GPL included in the
+** packaging of this file.  Please review the following information to
+** ensure the GNU General Public License version 3.0 requirements will be
+** met: http://www.gnu.org/copyleft/gpl.html.
+**
 ** If you have questions regarding the use of this file, please contact
 ** Nokia at qt-info@nokia.com.
-**
-**
-**
-**
-**
-**
-**
-**
 ** $QT_END_LICENSE$
 **
 ****************************************************************************/
@@ -44,7 +44,7 @@
 #include "propertydialog.h"
 
 #include <QtXml/QtXml>
-#include <private/qdbusutil_p.h>
+#include <../private/qdbusutil_p.h>
 
 class QDBusViewModel: public QDBusModel
 {
@@ -69,15 +69,18 @@ QDBusViewer::QDBusViewer(const QDBusConnection &connection, QWidget *parent)  :
     c(connection),
     objectPathRegExp(QLatin1String("\\[ObjectPath: (.*)\\]"))
 {
-    services = new QListWidget;
-    services->addItem("fooserv");
-    //services->setRootIsDecorated(false);
-    //services->setHeaderLabels(QStringList(QLatin1String("Services")));
+    servicesModel = new QStringListModel(this);
+    servicesFilterModel = new QSortFilterProxyModel(this);
+    servicesFilterModel->setSourceModel(servicesModel);
+    servicesFilterModel->setFilterCaseSensitivity(Qt::CaseInsensitive);
+    serviceFilterLine = new QLineEdit(this);
+    serviceFilterLine->setPlaceholderText(tr("Search..."));
+    servicesView = new QListView(this);
+    servicesView->setModel(servicesFilterModel);
 
-    tree = new QTreeView(parent);
-    tree->setAttribute(Qt::WA_Maemo5StackedWindow);
-    tree->setWindowFlags(tree->windowFlags() | Qt::Window);
+    connect(serviceFilterLine, SIGNAL(textChanged(QString)), servicesFilterModel, SLOT(setFilterFixedString(QString)));
 
+    tree = new QTreeView;
     tree->setContextMenuPolicy(Qt::CustomContextMenu);
 
     connect(tree, SIGNAL(activated(QModelIndex)), this, SLOT(activate(QModelIndex)));
@@ -90,22 +93,28 @@ QDBusViewer::QDBusViewer(const QDBusConnection &connection, QWidget *parent)  :
     QShortcut *refreshShortcut = new QShortcut(QKeySequence::Refresh, tree);
     connect(refreshShortcut, SIGNAL(activated()), this, SLOT(refreshChildren()));
 
-    QVBoxLayout *topLayout = new QVBoxLayout(this);
-    log = new QTextBrowser(tree);
-    log->setAttribute(Qt::WA_Maemo5StackedWindow);
-    log->setWindowFlags(tree->windowFlags() | Qt::Window);
+    QVBoxLayout *layout = new QVBoxLayout(this);
+    QSplitter *topSplitter = new QSplitter(Qt::Vertical, this);
+    layout->addWidget(topSplitter);
 
+    log = new QTextBrowser;
     connect(log, SIGNAL(anchorClicked(QUrl)), this, SLOT(anchorClicked(QUrl)));
 
-    QHBoxLayout *layout = new QHBoxLayout;
-    layout->addWidget(services, 1);
-    //layout->addWidget(tree, 2);
+    QSplitter *splitter = new QSplitter(topSplitter);
+    splitter->addWidget(servicesView);
 
-    topLayout->addLayout(layout);
-    //topLayout->addWidget(log);
+    QWidget *servicesWidget = new QWidget;
+    QVBoxLayout *servicesLayout = new QVBoxLayout(servicesWidget);
+    servicesLayout->addWidget(serviceFilterLine);
+    servicesLayout->addWidget(servicesView);
+    splitter->addWidget(servicesWidget);
+    splitter->addWidget(tree);
 
-    connect(services, SIGNAL(itemActivated(QListWidgetItem*)),
-            this, SLOT(serviceChanged(QListWidgetItem*)));
+    topSplitter->addWidget(splitter);
+    topSplitter->addWidget(log);
+
+    connect(servicesView->selectionModel(), SIGNAL(currentChanged(QModelIndex,QModelIndex)),
+            this, SLOT(serviceChanged(QModelIndex)));
     connect(tree, SIGNAL(customContextMenuRequested(QPoint)),
             this, SLOT(showContextMenu(QPoint)));
 
@@ -140,15 +149,11 @@ void QDBusViewer::logError(const QString &msg)
 
 void QDBusViewer::refresh()
 {
-    services->clear();
+    servicesModel->removeRows(0, servicesModel->rowCount());
 
     if (c.isConnected()) {
         const QStringList serviceNames = c.interface()->registeredServiceNames();
-        foreach (QString service, serviceNames) {
-            if (!service.startsWith(":")) {
-                new QListWidgetItem(service, services);
-            }
-        }
+        servicesModel->setStringList(serviceNames);
     }
 }
 
@@ -164,6 +169,7 @@ void QDBusViewer::activate(const QModelIndex &item)
     sig.mPath = model->dBusPath(item);
     sig.mInterface = model->dBusInterface(item);
     sig.mName = model->dBusMethodName(item);
+    sig.mTypeSig = model->dBusTypeSignature(item);
 
     switch (model->itemType(item)) {
     case QDBusModel::SignalItem:
@@ -180,7 +186,6 @@ void QDBusViewer::activate(const QModelIndex &item)
     }
 }
 
-
 void QDBusViewer::getProperty(const BusSignature &sig)
 {
     QDBusMessage message = QDBusMessage::createMethodCall(sig.mService, sig.mPath, QLatin1String("org.freedesktop.DBus.Properties"), QLatin1String("Get"));
@@ -218,6 +223,17 @@ void QDBusViewer::setProperty(const BusSignature &sig)
 
 }
 
+static QString getDbusSignature(const QMetaMethod& method)
+{
+    // create a D-Bus type signature from QMetaMethod's parameters
+    QString sig;
+    for (int i = 0; i < method.parameterTypes().count(); ++i) {
+        QVariant::Type type = QVariant::nameToType(method.parameterTypes().at(i));
+        sig.append(QString::fromLatin1(QDBusMetaType::typeToSignature(type)));
+    }
+    return sig;
+}
+
 void QDBusViewer::callMethod(const BusSignature &sig)
 {
     QDBusInterface iface(sig.mService, sig.mPath, sig.mInterface, c);
@@ -228,7 +244,8 @@ void QDBusViewer::callMethod(const BusSignature &sig)
     for (int i = 0; i < mo->methodCount(); ++i) {
         const QString signature = QString::fromLatin1(mo->method(i).signature());
         if (signature.startsWith(sig.mName) && signature.at(sig.mName.length()) == QLatin1Char('('))
-            method = mo->method(i);
+            if (getDbusSignature(mo->method(i)) == sig.mTypeSig)
+                method = mo->method(i);
     }
     if (!method.signature()) {
         QMessageBox::warning(this, tr("Unable to find method"),
@@ -288,6 +305,7 @@ void QDBusViewer::showContextMenu(const QPoint &point)
     sig.mPath = model->dBusPath(item);
     sig.mInterface = model->dBusInterface(item);
     sig.mName = model->dBusMethodName(item);
+    sig.mTypeSig = model->dBusTypeSignature(item);
 
     QMenu menu;
     menu.addAction(refreshAction);
@@ -390,18 +408,17 @@ void QDBusViewer::dumpMessage(const QDBusMessage &message)
     log->append(out);
 }
 
-void QDBusViewer::serviceChanged(QListWidgetItem *item)
+void QDBusViewer::serviceChanged(const QModelIndex &index)
 {
     delete tree->model();
 
     currentService.clear();
-    if (!item)
+    if (!index.isValid())
         return;
-    currentService = item->text();
+    currentService = index.data().toString();
 
     tree->setModel(new QDBusViewModel(currentService, c));
     connect(tree->model(), SIGNAL(busError(QString)), this, SLOT(logError(QString)));
-    tree->show();
 }
 
 void QDBusViewer::serviceRegistered(const QString &service)
@@ -409,34 +426,38 @@ void QDBusViewer::serviceRegistered(const QString &service)
     if (service == c.baseService())
         return;
 
-    new QListWidgetItem(service, services);
+    servicesModel->insertRows(0, 1);
+    servicesModel->setData(servicesModel->index(0, 0), service);
 }
 
-static QListWidgetItem *findItem(const QListWidget *services, const QString &name)
+static QModelIndex findItem(QStringListModel *servicesModel, const QString &name)
 {
-    for (int i = 0; i < services->count(); ++i) {
-        if (services->item(i)->text() == name)
-            return services->item(i);
-    }
-    return 0;
+    QModelIndexList hits = servicesModel->match(servicesModel->index(0, 0), Qt::DisplayRole, name);
+    if (hits.isEmpty())
+        return QModelIndex();
+
+    return hits.first();
 }
 
 void QDBusViewer::serviceUnregistered(const QString &name)
 {
-    delete findItem(services, name);
+    QModelIndex hit = findItem(servicesModel, name);
+    if (!hit.isValid())
+        return;
+    servicesModel->removeRows(hit.row(), 1);
 }
 
 void QDBusViewer::serviceOwnerChanged(const QString &name, const QString &oldOwner,
                                       const QString &newOwner)
 {
-    QListWidgetItem *item = findItem(services, name);
+    QModelIndex hit = findItem(servicesModel, name);
 
-    if (!item && oldOwner.isEmpty() && !newOwner.isEmpty())
+    if (!hit.isValid() && oldOwner.isEmpty() && !newOwner.isEmpty())
         serviceRegistered(name);
-    else if (item && !oldOwner.isEmpty() && newOwner.isEmpty())
-        delete item;
-    else if (item && !oldOwner.isEmpty() && !newOwner.isEmpty()) {
-        delete item;
+    else if (hit.isValid() && !oldOwner.isEmpty() && newOwner.isEmpty())
+        servicesModel->removeRows(hit.row(), 1);
+    else if (hit.isValid() && !oldOwner.isEmpty() && !newOwner.isEmpty()) {
+        servicesModel->removeRows(hit.row(), 1);
         serviceRegistered(name);
     }
 }
@@ -455,7 +476,7 @@ void QDBusViewer::about()
 
     box.setText(QString::fromLatin1("<center><img src=\":/trolltech/qdbusviewer/images/qdbusviewer-128.png\">"
                 "<h3>%1</h3>"
-                "<p>Version %2</p></center>"
+                "<p>Version %2 compiled for Maemo by Halftux</p></center>"
                 "<p>Copyright (C) 2010 Nokia Corporation and/or its subsidiary(-ies).</p>")
             .arg(tr("D-Bus Viewer")).arg(QLatin1String(QT_VERSION_STR)));
     box.setWindowTitle(tr("D-Bus Viewer"));
